import os
import sys

__all__ = ("LRU",)


NO_EXTENSIONS = bool(os.environ.get("TARINA_NO_EXTENSIONS"))  # type: bool
if sys.implementation.name != "cpython":
    NO_EXTENSIONS = True


if not NO_EXTENSIONS:  # pragma: no branch
    try:
        from ._lru_c import LRU as LRU  # type: ignore[misc]
    except Exception:  # pragma: no cover
        from ._lru_py import LRU as LRU  # type: ignore[misc]
else:
    from ._lru_py import LRU as LRU  # type: ignore[misc]



from __future__ import annotations

from collections import OrderedDict
from collections.abc import Hashable, Iterable
from typing import Any, Callable, Generic, TypeVar, overload

_KT = TypeVar("_KT", bound=Hashable)
_VT = TypeVar("_VT")
_T = TypeVar("_T")


class LRU(Generic[_KT, _VT]):
    __slots__ = ("__max", "__cache", "__size", "__callback")

    def __init__(self, size: int, callback: Callable[[_KT, _VT], Any] | None = None) -> None:
        if size < 1:
            raise ValueError("Size should be a positive number")
        self.__max = size
        self.__cache = OrderedDict()
        self.__size = 0
        self.__callback = callback

    def clear(self) -> None:
        self.__cache.clear()

    @overload
    def get(self, key: _KT) -> _VT | None: ...

    @overload
    def get(self, key: _KT, instead: _VT | _T) -> _VT | _T: ...

    def get(self, key: _KT, instead: _VT | _T | None = None) -> _VT | _T | None:
        if key in self.__cache:
            self.__cache.move_to_end(key, last=False)
            return self.__cache[key]
        return instead

    def get_size(self) -> int:
        return self.__max

    def has_key(self, key: _KT) -> bool:
        return key in self.__cache

    def keys(self) -> list[_KT]:
        return list(self.__cache.keys())

    def values(self) -> list[_VT]:
        return list(self.__cache.values())

    def items(self) -> list[tuple[_KT, _VT]]:
        return list(self.__cache.items())

    def peek_first_item(self) -> tuple[_KT, _VT] | None:
        return self.items()[0] if self.__cache else None

    def peek_last_item(self) -> tuple[_KT, _VT] | None:
        return self.items()[-1] if self.__cache else None

    @overload
    def pop(self, key: _KT) -> _VT | None: ...

    @overload
    def pop(self, key: _KT, default: _VT | _T) -> _VT | _T: ...

    def pop(self, key: _KT, default: _VT | _T | None = None) -> _VT | _T:
        return self.__cache.pop(key, default)

    def popitem(self, least_recent: bool = True) -> tuple[_KT, _VT]:
        return self.__cache.popitem(last=least_recent)

    @overload
    def setdefault(self: LRU[_KT, _T | None], key: _KT) -> _T | None: ...

    @overload
    def setdefault(self, key: _KT, default: _VT) -> _VT: ...

    def setdefault(self, key: _KT, default: _VT | None = None):
        if key in self.__cache:
            return self.__cache[key]
        self.__setitem__(key, default)  # type: ignore
        return default

    def set_callback(self, callback: Callable[[_KT, _VT], Any]) -> None:
        self.__callback = callback

    def set_size(self, size: int) -> None:
        self.__max = size
        if self.__max < self.__size:
            for _ in range(self.__size - self.__max):
                k, v = self.__cache.popitem(last=True)
                self.__size -= 1
                if self.__callback:
                    self.__callback(k, v)

    @overload
    def update(self, __m: Iterable[tuple[_KT, _VT]], **kwargs: _VT) -> None: ...

    @overload
    def update(self, **kwargs: _VT) -> None: ...

    def update(self, *args, **kwargs):
        self.__cache.update(*args, **kwargs)

    __contains__ = has_key

    def __delitem__(self, key: _KT) -> None:
        self.__cache.pop(key)

    def __getitem__(self, item: _KT) -> _VT:
        if item in self.__cache:
            self.__cache.move_to_end(item, last=False)
            return self.__cache[item]
        raise KeyError(item)

    def __len__(self) -> int:
        return len(self.__cache)

    def __repr__(self) -> str:
        return repr(self.__cache)

    def __setitem__(self, key: _KT, value: _VT) -> None:
        if key in self.__cache:
            self.__cache.move_to_end(key, last=False)
            self.__cache[key] = value
            return
        self.__cache[key] = value
        self.__cache.move_to_end(key, last=False)
        self.__size += 1
        if self.__max < self.__size:
            _k, _v = self.__cache.popitem(last=True)
            self.__size -= 1
            if self.__callback:
                self.__callback(_k, _v)



# cython: language_level=3, boundscheck=False, cdivision=True, wraparound=False, initializedcheck=False, infer_types=True, binding=True


from cpython.dict cimport PyDict_Contains, PyDict_GetItem
from cpython.int cimport PyInt_AS_LONG
from cpython.list cimport PyList_Append, PyList_GET_ITEM, PyList_GET_SIZE, PyList_Insert
from cpython.unicode cimport (
    PyUnicode_GET_LENGTH,
    PyUnicode_Join,
    PyUnicode_Split,
    PyUnicode_Substring,
    PyUnicode_Concat,
)


cdef extern from "Python.h":
    Py_UCS4 PyUnicode_READ_CHAR(object s, Py_ssize_t i)
    cdef Py_ssize_t PY_SSIZE_T_MAX


cdef extern from "_op.h":
    bint str_contains(object str_obj, Py_UCS4 ch)
    unicode str_strip(object str_obj, int striptype, object chars_obj)
    cdef int LEFTSTRIP
    cdef int RIGHTSTRIP
    cdef int BOTHSTRIP

cdef dict QUOTES = {'"': '"', "'": "'"}
cdef unicode CRLF = "\n\r"

def split(str text, str separator, bint crlf=True):
    if crlf:
        separator = PyUnicode_Concat(separator, CRLF)
    text = str_strip(text, BOTHSTRIP, separator)
    cdef:
        bint escape = 0
        list result = []
        Py_UCS4 quotation = 0
        Py_UCS4 ch = 0
        Py_ssize_t index = 0
        Py_ssize_t length = PyUnicode_GET_LENGTH(text)
        list quoted_sep_index = []
        Py_ssize_t last_sep_index = 0
        Py_ssize_t last_quote_index = 0
        Py_ssize_t _len
        Py_ssize_t i = 0

    while index < length:
        ch = PyUnicode_READ_CHAR(text, index)
        index += 1
        if PyDict_Contains(QUOTES, ch):
            if index == 1 + escape + max(last_sep_index, last_quote_index) and quotation == 0:
                quotation = PyUnicode_READ_CHAR(<str>PyDict_GetItem(QUOTES, ch), 0)
            elif (PyList_GET_SIZE(result) == 0 or str_contains(separator, <Py_UCS4>(<object>PyList_GET_ITEM(result, PyList_GET_SIZE(result)-1))) == 0) and ch == quotation:
                quotation = 0
                last_quote_index = index
            else:
                PyList_Append(result, ch)
            if escape:
                result[PyList_GET_SIZE(result)-1] = ch
        elif str_contains(separator, ch):
            if quotation:
                PyList_Append(quoted_sep_index, PyList_GET_SIZE(result) + 1)
                PyList_Append(result, ch)
            else:
                last_sep_index = index
                if result and (<object>PyList_GET_ITEM(result, PyList_GET_SIZE(result)-1)) != '\1':
                    PyList_Append(result, '\1')
        else:
            PyList_Append(result, ch)
        escape = ch == 92
    if PyList_GET_SIZE(result) == 0:
        return []
    if quotation and PyList_GET_SIZE(quoted_sep_index):
        PyList_Insert(result, last_sep_index, PyUnicode_READ_CHAR(text,  last_quote_index or last_sep_index))
        # result[first_quoted_sep_index] = '\1'
        _len = PyList_GET_SIZE(quoted_sep_index)
        while i < _len:
            result[PyInt_AS_LONG(<object>PyList_GET_ITEM(quoted_sep_index, i))] = '\1'
            i += 1
    return PyUnicode_Split(PyUnicode_Join('', result), '\1', -1)


def split_once(str text, str separator, bint crlf=True):
    if crlf:
        separator = PyUnicode_Concat(separator, CRLF)
    text = str_strip(text, LEFTSTRIP, separator)
    cdef:
        Py_ssize_t index = 0
        list out_text = []
        Py_UCS4 quotation = 0
        Py_UCS4 ch = 0
        bint escape = 0
        bint sep = 0
        Py_ssize_t length = PyUnicode_GET_LENGTH(text)
        Py_ssize_t first_quoted_sep_index = -1
        Py_ssize_t last_quote_index = 0
    while index < length:
        ch = PyUnicode_READ_CHAR(text, index)
        index += 1
        if str_contains(separator, ch):
            if quotation == 0:
                sep = 1
                continue
            if first_quoted_sep_index == -1:
                first_quoted_sep_index = index
        if sep == 1:
            index -= 1
            break
        if PyDict_Contains(QUOTES, ch):  # 遇到引号括起来的部分跳过分隔
            if index == 1 + escape + last_quote_index and quotation == 0:
                quotation = PyUnicode_READ_CHAR(<str>PyDict_GetItem(QUOTES, ch), 0)
            elif str_contains(separator, PyUnicode_READ_CHAR(text, index-2)) == 0 and ch == quotation:
                last_quote_index = index
                first_quoted_sep_index = -1
                quotation = 0
            else:
                PyList_Append(out_text, ch)
            if escape:
                out_text[PyList_GET_SIZE(out_text)-1] = ch
            escape = 0
        else:
            PyList_Append(out_text, ch)
        escape = ch == 92
    if index == length:
        if first_quoted_sep_index == -1:
            return PyUnicode_Join('', out_text) if last_quote_index else str_strip(text, RIGHTSTRIP, separator), ''
        return PyUnicode_Substring(text, 0, first_quoted_sep_index-1), PyUnicode_Substring(text, first_quoted_sep_index, PY_SSIZE_T_MAX)
    return PyUnicode_Join('', out_text), PyUnicode_Substring(text, index, PY_SSIZE_T_MAX)


def split_once_without_escape(str text, str separator, bint crlf=True):
    if crlf:
        separator = PyUnicode_Concat(separator, CRLF)
    text = str_strip(text, LEFTSTRIP, separator)
    cdef:
        Py_ssize_t index = 0
        Py_UCS4 quotation = 0
        Py_UCS4 ch = 0
        Py_ssize_t length = PyUnicode_GET_LENGTH(text)
        Py_ssize_t first_quoted_sep_index = -1
        Py_ssize_t last_quote_index = 0
    while index < length:
        ch = PyUnicode_READ_CHAR(text, index)
        index += 1
        if str_contains(separator, ch):
            if quotation == 0:
                break
            if first_quoted_sep_index == -1:
                first_quoted_sep_index = index
        if PyDict_Contains(QUOTES, ch):  # 遇到引号括起来的部分跳过分隔
            if index == 1 + last_quote_index and quotation == 0:
                quotation = PyUnicode_READ_CHAR(<str>PyDict_GetItem(QUOTES, ch), 0)
            elif str_contains(separator, PyUnicode_READ_CHAR(text, index-2)) == 0 and ch == quotation:
                last_quote_index = index
                first_quoted_sep_index = -1
                quotation = 0
    if index == length:
        if first_quoted_sep_index == -1:
            return PyUnicode_Substring(text, 0, last_quote_index) if last_quote_index else str_strip(text, RIGHTSTRIP, separator), ''
        return PyUnicode_Substring(text, 0, first_quoted_sep_index-1), PyUnicode_Substring(text, first_quoted_sep_index, PY_SSIZE_T_MAX)
    return PyUnicode_Substring(text, 0, index-1), PyUnicode_Substring(text, index, PY_SSIZE_T_MAX)


cpdef inline tuple split_once_index_only(str text, str separator, Py_ssize_t offset, bint crlf=True):
    if crlf:
        separator = PyUnicode_Concat(separator, CRLF)
    cdef:
        Py_ssize_t index = offset
        Py_UCS4 quotation = 0
        Py_UCS4 ch = 0
        Py_ssize_t sep = 0
        Py_ssize_t length = PyUnicode_GET_LENGTH(text)
        Py_ssize_t quoted_sep_index = -1
        Py_ssize_t quoted_sep = 0
        Py_ssize_t last_quote_index = 0
    while index < length:
        ch = PyUnicode_READ_CHAR(text, index)
        index += 1
        if str_contains(separator, ch):
            if quotation == 0:
                sep += 1
                continue
            quoted_sep_index = index
            quoted_sep += 1
        if sep:
            index -= 1
            break
        if PyDict_Contains(QUOTES, ch):  # 遇到引号括起来的部分跳过分隔
            if index == 1 + (last_quote_index or offset) and quotation == 0:
                quotation = PyUnicode_READ_CHAR(<str>PyDict_GetItem(QUOTES, ch), 0)
            elif str_contains(separator, PyUnicode_READ_CHAR(text, index-2)) == 0 and ch == quotation:
                last_quote_index = index
                quoted_sep_index = -1
                quoted_sep = 0
                quotation = 0
    if index == length and quoted_sep_index != -1:
        return quoted_sep_index, quoted_sep
    return index, sep


cdef class String:
    cdef public Py_ssize_t left_index
    cdef public Py_ssize_t next_index
    cdef public Py_ssize_t offset
    cdef public Py_ssize_t len
    cdef public str text

    def __init__(self, str text):
        self.text = text
        self.len = PyUnicode_GET_LENGTH(text)
        self.left_index = 0
        self.offset = 0
        self.next_index = 0

    def step(self, str separator, bint crlf=True):
        self.next_index, self.offset = split_once_index_only(self.text, separator, self.left_index, crlf)

    def val(self):
        return PyUnicode_Substring(self.text, self.left_index, self.next_index - self.offset)

    def apply(self, int left = -1):
        if left == -1:
            self.left_index = self.next_index
        else:
            self.next_index = self.left_index = left
        self.offset = 0

    def rest(self):
        return PyUnicode_Substring(self.text, self.left_index, self.len)

    @property
    def complete(self):
        return self.left_index == self.len

    @property
    def will_complete(self):
        return self.next_index == self.len

    def __repr__(self):
        return f"String({self.text!r}[{self.left_index}:{self.next_index - self.offset}])"

    def __str__(self):
        return self.val()



import re
from datetime import datetime, timedelta, timezone
from typing import Union

MILLI_SECOND = 1
SECOND = MILLI_SECOND * 1000
MINUTE = SECOND * 60
HOUR = MINUTE * 60
DAY = HOUR * 24
WEEK = DAY * 7

NUMERIC = r"\d+(?:\.\d+)?"
TIME_REGEXP = re.compile(
    "^"
    + "".join(
        f"(?:(?:{unit}({NUMERIC}))|(?:({NUMERIC}){unit}))?"
        for unit in [
            "w(?:eek(?:s)?)?",
            "d(?:ay(?:s)?)?",
            "h(?:our(?:s)?)?",
            "m(?:in(?:ute)?(?:s)?)?",
            "s(?:ec(?:ond)?(?:s)?)?",
        ]
    )
)


def get_total_seconds(delta: datetime):
    offset = delta.utcoffset()
    if not offset:
        return 0
    return offset.total_seconds() / 60


class DateParser:
    timezone_offset = get_total_seconds(datetime.now(timezone.utc).astimezone())

    @classmethod
    def set_timezone_offset(cls, offset: float):
        cls.timezone_offset = offset

    @classmethod
    def get_timezone_offset(cls):
        return cls.timezone_offset

    @classmethod
    def get_date_number(cls, date: Union[float, datetime, None] = None, offset: Union[float, None] = None):
        if date is None:
            date = datetime.now()
        if isinstance(date, float):
            date = datetime.fromtimestamp(date)
        if offset is None:
            offset = cls.timezone_offset
        return int((date.timestamp() / MINUTE - offset) / 1440)  # type: ignore

    @classmethod
    def from_date_number(cls, value: float, offset: Union[float, None] = None):
        date = datetime.fromtimestamp(value * DAY)
        if offset is None:
            offset = cls.timezone_offset
        return datetime.fromtimestamp(date.timestamp() + offset * MINUTE)

    @classmethod
    def parse(cls, pattern: str):
        if (capture := TIME_REGEXP.match(pattern)) and capture.groups():
            if stamp := (
                float(capture[1] or capture[2] or 0) * WEEK
                + float(capture[3] or capture[4] or 0) * DAY
                + float(capture[5] or capture[6] or 0) * HOUR
                + float(capture[7] or capture[8] or 0) * MINUTE
                + float(capture[9] or capture[10] or 0) * SECOND
            ):
                return datetime.now() + timedelta(milliseconds=stamp)
        if re.match(r"^\d{1,2}(:\d{1,2}){1,2}$", pattern):
            return datetime.fromisoformat(f"{datetime.now().strftime('%Y-%m-%d')}-{pattern}")
        if re.match(r"^\d{1,2}-\d{1,2}-\d{1,2}(:\d{1,2}){1,2}$", pattern):
            return datetime.fromisoformat(f"{datetime.now().year}-{pattern}")
        return datetime.fromisoformat(pattern)

    @classmethod
    def format(cls, ms: float):
        abs_ms = abs(ms)
        if abs_ms >= DAY - HOUR / 2:
            return f"{round(ms / DAY)}d"
        elif abs_ms >= HOUR - MINUTE / 2:
            return f"{round(ms / HOUR)}h"
        elif abs_ms >= MINUTE - SECOND / 2:
            return f"{round(ms / MINUTE)}m"
        elif abs_ms >= SECOND:
            return f"{round(ms / SECOND)}s"
        return f"{ms}ms"

    @classmethod
    def to_digits(cls, source: int, length: int = 2):
        return str(source).zfill(length)

    @classmethod
    def template(cls, template: str, time: Union[datetime, None] = None):
        if time is None:
            time = datetime.now()
        return (
            template.replace("yyyy", str(time.year))
            .replace("yy", str(time.year)[2:])
            .replace("MM", cls.to_digits(time.month))
            .replace("dd", cls.to_digits(time.day))
            .replace("hh", cls.to_digits(time.hour))
            .replace("mm", cls.to_digits(time.minute))
            .replace("ss", cls.to_digits(time.second))
            .replace("SSS", cls.to_digits(time.microsecond, 3))
        )



import keyword
import re
from pathlib import Path
from typing import Optional

from tarina.lang import lang
from tarina.lang.schema import _Subtypes, _TemplateDict, get_template
from tarina.tools import pascal_case


class LangItem:
    def __init__(self, scope: str, type: str):
        self.scope = scope
        self.type = type

    def set(self, content: str, locale: Optional[str] = None):
        lang.set(self.scope, self.type, content, locale)

    def __call__(self, locale: Optional[str] = None, **format_kwargs):
        return lang.require(self.scope, self.type, locale).format(**format_kwargs)

    def __iter__(self):
        return iter([self.scope, self.type])


class LangModel:
    pass


MODEL_TEMPLATE = """\
# This file is @generated by tarina.lang CLI tool
# It is not intended for manual editing.

from tarina.lang.model import LangItem, LangModel


{scope_classes}class Lang(LangModel):
{scopes}\
"""

SCOPE_TEMPLATE = """\
class {scope}:
{types}

"""

TYPE_TEMPLATE = """\
    {name}: LangItem = LangItem("{scope}", "{type}")
"""


def generate_model(root: Path):
    template = get_template(root)
    if "scopes" not in template:
        raise KeyError("Template file must have a 'scopes' key")
    scopes: list[_TemplateDict] = template["scopes"]
    scopes_classes = ""
    scopes_str = ""
    for s in scopes:
        scope = re.sub(r"[\W\s]+", "_", s["scope"])
        if scope in keyword.kwlist:
            scope += "_"
        scope_class = pascal_case(scope)
        if scope_class == "Lang":
            scope_class += "_"

        def visit_types(_scope: str, _types: "list[str | _Subtypes]", prefix: str = ""):

            types_str = ""
            for t in _types:
                if isinstance(t, dict):
                    subtype = re.sub(r"[\W\s]+", "_", t["subtype"])
                    if subtype in keyword.kwlist:
                        subtype += "_"
                    yield from visit_types(
                        f"{_scope}{pascal_case(subtype)}", t["types"], prefix=f"{prefix}{t['subtype']}."
                    )
                    types_str += f"    {subtype} = {_scope}{pascal_case(subtype)}\n"
                else:
                    name = re.sub(r"[\W\s]+", "_", t)
                    if name in keyword.kwlist:
                        name += "_"
                    types_str += TYPE_TEMPLATE.format(name=name, scope=s["scope"], type=f"{prefix}{t}")
            yield SCOPE_TEMPLATE.format(scope=_scope, types=types_str)

        scopes_classes += "\n".join(visit_types(scope_class, s["types"]))
        scopes_str += f"    {scope} = {scope_class}\n"
    return MODEL_TEMPLATE.format(scope_classes=scopes_classes, scopes=scopes_str)


def write_model(root: Path):
    model = generate_model(root)
    with (root / "model.py").open("w", encoding="utf-8") as f:
        f.write(model)
